import logging
from abc import ABC, abstractmethod
from textwrap import TextWrapper
from typing import List, Tuple

import curlify

from pricehist import exceptions
from pricehist.series import Series


class BaseSource(ABC):
    @abstractmethod
    def id(self) -> str:
        pass  # pragma: nocover

    @abstractmethod
    def name(self) -> str:
        pass  # pragma: nocover

    @abstractmethod
    def description(self) -> str:
        pass  # pragma: nocover

    @abstractmethod
    def source_url(self) -> str:
        pass  # pragma: nocover

    @abstractmethod
    def start(self) -> str:
        pass  # pragma: nocover

    @abstractmethod
    def types(self) -> List[str]:
        pass  # pragma: nocover

    @abstractmethod
    def notes(self) -> str:
        pass  # pragma: nocover

    def normalizesymbol(self, str) -> str:
        return str.upper()

    @abstractmethod
    def symbols(self) -> List[Tuple[str, str]]:
        pass  # pragma: nocover

    def search(self, query) -> List[Tuple[str, str]]:
        pass  # pragma: nocover

    @abstractmethod
    def fetch(self, series: Series) -> Series:
        pass  # pragma: nocover

    def log_curl(self, response):
        curl = curlify.to_curl(response.request, compressed=True)
        logging.debug(curl)
        return response

    def format_symbols(self) -> str:
        with exceptions.handler():
            symbols = self.symbols()

        width = max([len(sym) for sym, desc in symbols] + [0])
        lines = [sym.ljust(width + 4) + desc + "\n" for sym, desc in symbols]
        return "".join(lines)

    def format_search(self, query) -> str:
        with exceptions.handler():
            symbols = self.search(query)

        if symbols is None:
            logging.error(f"Symbol search is not possible for the {self.id()} source.")
            exit(1)
        elif symbols == []:
            logging.info(f"No results found for query '{query}'.")
            return ""
        else:
            width = max([len(sym) for sym, desc in symbols] + [0])
            lines = [sym.ljust(width + 4) + desc + "\n" for sym, desc in symbols]
            return "".join(lines)

    def format_info(self, total_width=80) -> str:
        k_width = 11
        with exceptions.handler():
            parts = [
                self._fmt_field("ID", self.id(), k_width, total_width),
                self._fmt_field("Name", self.name(), k_width, total_width),
                self._fmt_field(
                    "Description", self.description(), k_width, total_width
                ),
                self._fmt_field("URL", self.source_url(), k_width, total_width, False),
                self._fmt_field("Start", self.start(), k_width, total_width),
                self._fmt_field("Types", ", ".join(self.types()), k_width, total_width),
                self._fmt_field("Notes", self.notes(), k_width, total_width),
            ]
        return "\n".join(filter(None, parts))

    def _fmt_field(self, key, value, key_width, total_width, force=True):
        separator = " : "
        initial_indent = key + (" " * (key_width - len(key))) + separator
        subsequent_indent = " " * len(initial_indent)
        wrapper = TextWrapper(
            width=total_width,
            drop_whitespace=True,
            initial_indent=initial_indent,
            subsequent_indent=subsequent_indent,
            break_long_words=force,
        )
        first, *rest = value.split("\n")
        first_output = wrapper.wrap(first)
        wrapper.initial_indent = subsequent_indent
        rest_output = sum([wrapper.wrap(line) if line else [""] for line in rest], [])
        output = "\n".join(first_output + rest_output)
        if output != "":
            return output
        else:
            return None



import dataclasses
import json
from decimal import Decimal

import requests

from pricehist import exceptions
from pricehist.price import Price

from .basesource import BaseSource


class BankOfCanada(BaseSource):
    def id(self):
        return "bankofcanada"

    def name(self):
        return "Bank of Canada"

    def description(self):
        return "Daily exchange rates of the Canadian dollar from the Bank of Canada"

    def source_url(self):
        return "https://www.bankofcanada.ca/valet/docs"

    def start(self):
        return "2017-01-03"

    def types(self):
        return ["default"]

    def notes(self):
        return (
            "Currently, only daily exchange rates are supported. They are "
            "published once each business day by 16:30 ET. "
            "All Bank of Canada exchange rates are indicative rates only.\n"
            "To request support for other data provided by the "
            "Bank of Canada Valet Web Services, please open an "
            "issue in pricehist's Gitlab project. "
        )

    def symbols(self):
        url = "https://www.bankofcanada.ca/valet/lists/series/json"

        try:
            response = self.log_curl(requests.get(url))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            data = json.loads(response.content)
            series_names = data["series"].keys()
            fx_series_names = [
                n for n in series_names if len(n) == 8 and n[0:2] == "FX"
            ]
            results = [
                (f"{n[2:5]}/{n[5:9]}", data["series"][n]["description"])
                for n in sorted(fx_series_names)
            ]

        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if not results:
            raise exceptions.ResponseParsingError("Expected data not found")
        else:
            return results

    def fetch(self, series):
        if len(series.base) != 3 or len(series.quote) != 3:
            raise exceptions.InvalidPair(series.base, series.quote, self)

        series_name = f"FX{series.base}{series.quote}"
        data = self._data(series, series_name)

        prices = []
        for o in data.get("observations", []):
            prices.append(Price(o["d"], Decimal(o[series_name]["v"])))

        return dataclasses.replace(series, prices=prices)

    def _data(self, series, series_name):
        url = f"https://www.bankofcanada.ca/valet/observations/{series_name}/json"
        params = {
            "start_date": series.start,
            "end_date": series.end,
            "order_dir": "asc",
        }

        try:
            response = self.log_curl(requests.get(url, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        code = response.status_code
        text = response.text

        try:
            result = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if code == 404 and "not found" in text:
            raise exceptions.InvalidPair(series.base, series.quote, self)
        elif code == 400 and "End date must be greater than the Start date" in text:
            raise exceptions.BadResponse(result["message"])
        else:
            try:
                response.raise_for_status()
            except Exception as e:
                raise exceptions.BadResponse(str(e)) from e

        return result



from abc import ABC, abstractmethod

from pricehist.format import Format
from pricehist.series import Series
from pricehist.sources.basesource import BaseSource


class BaseOutput(ABC):
    @abstractmethod
    def format(self, series: Series, source: BaseSource, fmt: Format) -> str:
        pass  # pragma: nocover



from .beancount import Beancount
from .csv import CSV
from .gnucashsql import GnuCashSQL
from .json import JSON
from .ledger import Ledger

default = "csv"

by_type = {
    "beancount": Beancount(),
    "csv": CSV(),
    "json": JSON(),
    "jsonl": JSON(jsonl=True),
    "gnucash-sql": GnuCashSQL(),
    "ledger": Ledger(),
}



import csv
import dataclasses
import json
import logging
from datetime import datetime, timezone
from decimal import Decimal

import requests

from pricehist import __version__, exceptions
from pricehist.price import Price

from .basesource import BaseSource


class Yahoo(BaseSource):
    def id(self):
        return "yahoo"

    def name(self):
        return "Yahoo! Finance"

    def description(self):
        return (
            "Historical data for most Yahoo! Finance symbols, "
            "as available on the web page"
        )

    def source_url(self):
        return "https://finance.yahoo.com/"

    def start(self):
        # The "Download historical data in Yahoo Finance" page says
        # "Historical prices usually don't go back earlier than 1970", but
        # several do. Examples going back to 1962-01-02 include ED and IBM.
        return "1962-01-02"

    def types(self):
        return ["adjclose", "open", "high", "low", "close", "mid"]

    def notes(self):
        return (
            "Yahoo! Finance decommissioned its historical data API in 2017 but "
            "some historical data is available via its web page, as described in: "
            "https://help.yahoo.com/kb/"
            "download-historical-data-yahoo-finance-sln2311.html\n"
            f"{self._symbols_message()}\n"
            "In output the base and quote will be the Yahoo! symbol and its "
            "corresponding currency. Some symbols include the name of the quote "
            "currency (e.g. BTC-USD), so you may wish to use --fmt-base to "
            "remove the redundant information.\n"
            "When a symbol's historical data is unavilable due to data licensing "
            "restrictions, its web page will show no download button and "
            "pricehist will only find the current day's price."
        )

    def _symbols_message(self):
        return (
            "Find the symbol of interest on https://finance.yahoo.com/ and use "
            "that as the PAIR in your pricehist command. Prices for each symbol "
            "are quoted in its native currency."
        )

    def symbols(self):
        logging.info(self._symbols_message())
        return []

    def fetch(self, series):
        if series.quote:
            raise exceptions.InvalidPair(
                series.base, series.quote, self, "Don't specify the quote currency."
            )

        quote, history = self._data(series)

        prices = [
            Price(row["date"], amount)
            for row in history
            if (amount := self._amount(row, series.type))
        ]

        return dataclasses.replace(series, quote=quote, prices=prices)

    def _amount(self, row, type):
        if type == "mid" and row["high"] != "null" and row["low"] != "null":
            return sum([Decimal(row["high"]), Decimal(row["low"])]) / 2
        elif row[type] != "null":
            return Decimal(row[type])
        else:
            return None

    def _data(self, series) -> (dict, csv.DictReader):
        base_url = "https://query1.finance.yahoo.com/v7/finance"
        headers = {"User-Agent": f"pricehist/{__version__}"}

        spark_url = f"{base_url}/spark"
        spark_params = {
            "symbols": series.base,
            "range": "1d",
            "interval": "1d",
            "indicators": "close",
            "includeTimestamps": "false",
            "includePrePost": "false",
        }
        try:
            spark_response = self.log_curl(
                requests.get(spark_url, params=spark_params, headers=headers)
            )
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        code = spark_response.status_code
        text = spark_response.text
        if code == 404 and "No data found for spark symbols" in text:
            raise exceptions.InvalidPair(
                series.base, series.quote, self, "Symbol not found."
            )

        try:
            spark_response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            spark = json.loads(spark_response.content)
            quote = spark["spark"]["result"][0]["response"][0]["meta"]["currency"]
        except Exception as e:
            raise exceptions.ResponseParsingError(
                "The spark data couldn't be parsed. "
            ) from e

        start_ts = int(
            datetime.strptime(series.start, "%Y-%m-%d")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        )
        end_ts = int(
            datetime.strptime(series.end, "%Y-%m-%d")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        ) + (
            24 * 60 * 60
        )  # round up to include the last day

        history_url = f"{base_url}/download/{series.base}"
        history_params = {
            "period1": start_ts,
            "period2": end_ts,
            "interval": "1d",
            "events": "history",
            "includeAdjustedClose": "true",
        }

        try:
            history_response = self.log_curl(
                requests.get(history_url, params=history_params, headers=headers)
            )
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        code = history_response.status_code
        text = history_response.text

        if code == 404 and "No data found, symbol may be delisted" in text:
            raise exceptions.InvalidPair(
                series.base, series.quote, self, "Symbol not found."
            )
        if code == 400 and "Data doesn't exist" in text:
            raise exceptions.BadResponse(
                "No data for the given interval. Try requesting a larger interval."
            )

        elif code == 404 and "Timestamp data missing" in text:
            raise exceptions.BadResponse(
                "Data missing. The given interval may be for a gap in the data "
                "such as a weekend or holiday. Try requesting a larger interval."
            )

        try:
            history_response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            history_lines = history_response.content.decode("utf-8").splitlines()
            history_lines[0] = history_lines[0].lower().replace(" ", "")
            history = csv.DictReader(history_lines, delimiter=",")
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if history_lines[0] != "date,open,high,low,close,adjclose,volume":
            raise exceptions.ResponseParsingError("Unexpected CSV format")

        return (quote, history)



"""
ISO 4217 Currency data

Provides `ISO 4217 <https://www.iso.org/iso-4217-currency-codes.html>`_
currency data in a ready-to-use format, indexed by currency code. Historical
currencies are included and countries with no universal currency are ignored.

The data is read from vendored copies of the XML files published by the
maintainers of the standard:

* :file:`list-one.xml` (current currencies & funds)
* :file:`list-three.xml` (historical currencies & funds)

Classes:

    ISOCurrency

Functions:

    current_data_date() -> str
    historical_data_date() -> str
    by_code() -> dict[str, ISOCurrency]

"""

from dataclasses import dataclass, field
from importlib.resources import read_binary
from typing import List

from lxml import etree


@dataclass(frozen=False)
class ISOCurrency:
    code: str = None
    number: int = None
    minor_units: int = None
    name: str = None
    is_fund: bool = False
    countries: List[str] = field(default_factory=list)
    historical: bool = False
    withdrawal_date: str = None


def current_data_date():
    one = etree.fromstring(read_binary("pricehist.resources", "list-one.xml"))
    return one.cssselect("ISO_4217")[0].attrib["Pblshd"]


def historical_data_date():
    three = etree.fromstring(read_binary("pricehist.resources", "list-three.xml"))
    return three.cssselect("ISO_4217")[0].attrib["Pblshd"]


def by_code():
    result = {}

    one = etree.fromstring(read_binary("pricehist.resources", "list-one.xml"))
    three = etree.fromstring(read_binary("pricehist.resources", "list-three.xml"))

    for entry in three.cssselect("HstrcCcyNtry") + one.cssselect("CcyNtry"):
        if currency := _parse(entry):
            if existing := result.get(currency.code):
                existing.code = currency.code
                existing.number = currency.number
                existing.minor_units = currency.minor_units
                existing.name = currency.name
                existing.is_fund = currency.is_fund
                existing.countries += currency.countries
                existing.historical = currency.historical
                existing.withdrawal_date = currency.withdrawal_date
            else:
                result[currency.code] = currency

    return result


def _parse(entry):
    try:
        code = entry.cssselect("Ccy")[0].text
    except IndexError:
        return None  # Ignore countries without a universal currency

    try:
        number = int(entry.cssselect("CcyNbr")[0].text)
    except (IndexError, ValueError):
        number = None

    try:
        minor_units = int(entry.cssselect("CcyMnrUnts")[0].text)
    except (IndexError, ValueError):
        minor_units = None

    name = None
    is_fund = None
    if name_tags := entry.cssselect("CcyNm"):
        name = name_tags[0].text
        is_fund = name_tags[0].attrib.get("IsFund", "").upper() in ["TRUE", "WAHR"]

    countries = [t.text for t in entry.cssselect("CtryNm")]

    try:
        withdrawal_date = entry.cssselect("WthdrwlDt")[0].text
        historical = True
    except IndexError:
        withdrawal_date = None
        historical = False

    return ISOCurrency(
        code=code,
        number=number,
        minor_units=minor_units,
        name=name,
        is_fund=is_fund,
        countries=countries,
        historical=historical,
        withdrawal_date=withdrawal_date,
    )



import logging
import sys
from contextlib import contextmanager


@contextmanager
def handler():
    try:
        yield
    except SourceError as e:
        logging.debug("Critical exception encountered", exc_info=e)
        logging.critical(str(e))
        sys.exit(1)


class SourceError(Exception):
    """Base exception for errors rased by sources"""


class InvalidPair(SourceError, ValueError):
    """An invalid pair was requested."""

    def __init__(self, base, quote, source, message=None):
        self.base = base
        self.quote = quote
        self.source = source
        pair = "/".join([s for s in [base, quote] if s])
        insert = message + " " if message else ""

        full_message = (
            f"Invalid pair '{pair}'. {insert}"
            f"Run 'pricehist source {source.id()} --symbols' "
            f"for information about valid pairs."
        )
        super(InvalidPair, self).__init__(full_message)


class InvalidType(SourceError, ValueError):
    """An invalid price type was requested."""

    def __init__(self, type, base, quote, source):
        self.type = type
        self.pair = "/".join([s for s in [base, quote] if s])
        message = (
            f"Invalid price type '{type}' for pair '{self.pair}'. "
            f"Run 'pricehist source {source.id()}' "
            f"for information about valid types."
        )
        super(InvalidType, self).__init__(message)


class CredentialsError(SourceError):
    """Access credentials are unavailable or invalid."""

    def __init__(self, keys, source, msg=""):
        self.keys = keys
        self.source = source
        message = (
            f"Access credentials for source '{source.id()}' are unavailable "
            f"""or invalid. Set the environment variables '{"', '".join(keys)}' """
            f"correctly. Run 'pricehist source {source.id()}' for more "
            f"information about credentials."
        )
        if msg:
            message += f" {msg}"
        super(CredentialsError, self).__init__(message)


class RateLimit(SourceError):
    """Source request rate limit reached."""

    def __init__(self, message):
        super(RateLimit, self).__init__(f"{self.__doc__} {message}")


class RequestError(SourceError):
    """An error occured while making a request to the source."""

    def __init__(self, message):
        super(RequestError, self).__init__(f"{self.__doc__} {message}")


class BadResponse(SourceError):
    """A bad response was received from the source."""

    def __init__(self, message):
        super(BadResponse, self).__init__(f"{self.__doc__} {message}")


class ResponseParsingError(SourceError):
    """An error occurred while parsing data from the source."""

    def __init__(self, message):
        super(ResponseParsingError, self).__init__(f"{self.__doc__} {message}")



import logging
import sys


class Formatter(logging.Formatter):
    def format(self, record):
        s = record.msg % record.args if record.args else record.msg

        if record.exc_info:
            record.exc_text = self.formatException(record.exc_info)
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + "\n".join([f"  {line}" for line in record.exc_text.splitlines()])

        if record.levelno != logging.INFO:
            s = "\n".join([f"{record.levelname} {line}" for line in s.splitlines()])

        return s


def init():
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(Formatter())
    logging.root.addHandler(handler)
    logging.root.setLevel(logging.INFO)
    logging.getLogger("charset_normalizer").disabled = True


def show_debug():
    logging.root.setLevel(logging.DEBUG)



from .alphavantage import AlphaVantage
from .bankofcanada import BankOfCanada
from .coinbasepro import CoinbasePro
from .coindesk import CoinDesk
from .coinmarketcap import CoinMarketCap
from .ecb import ECB
from .exchangeratehost import ExchangeRateHost
from .yahoo import Yahoo

by_id = {
    source.id(): source
    for source in [
        AlphaVantage(),
        BankOfCanada(),
        CoinbasePro(),
        CoinDesk(),
        CoinMarketCap(),
        ECB(),
        ExchangeRateHost(),
        Yahoo(),
    ]
}


def formatted():
    width = max([len(k) for k, v in by_id.items()])
    lines = [k.ljust(width + 4) + v.name() for k, v in by_id.items()]
    return "\n".join(lines)



import dataclasses
import json
import logging
from decimal import Decimal

import requests

from pricehist import exceptions
from pricehist.price import Price

from .basesource import BaseSource


class CoinDesk(BaseSource):
    def id(self):
        return "coindesk"

    def name(self):
        return "CoinDesk Bitcoin Price Index"

    def description(self):
        return (
            "WARNING: This source is deprecated. Data stops at 2022-07-10.\n"
            "The documentation URL now redirects to the main page.\n"
            "An average of Bitcoin prices across leading global exchanges.\n"
            "Powered by CoinDesk, https://www.coindesk.com/price/bitcoin"
        )

    def source_url(self):
        return "https://www.coindesk.com/coindesk-api"

    def start(self):
        return "2010-07-17"

    def types(self):
        return ["close"]

    def notes(self):
        return ""

    def symbols(self):
        url = "https://api.coindesk.com/v1/bpi/supported-currencies.json"

        try:
            response = self.log_curl(requests.get(url))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            data = json.loads(response.content)
            relevant = [i for i in data if i["currency"] not in ["BTC", "XBT"]]
            results = [
                (f"BTC/{i['currency']}", f"Bitcoin against {i['country']}")
                for i in sorted(relevant, key=lambda i: i["currency"])
            ]
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if not results:
            raise exceptions.ResponseParsingError("Expected data not found")
        else:
            return results

    def fetch(self, series):
        logging.warning("This source is deprecated. Data stops at 2022-07-10.")

        if series.base != "BTC" or series.quote in ["BTC", "XBT"]:
            # BTC is the only valid base.
            # BTC as the quote will return BTC/USD, which we don't want.
            # XBT as the quote will fail with HTTP status 500.
            raise exceptions.InvalidPair(series.base, series.quote, self)

        data = self._data(series)

        prices = []
        for (d, v) in data.get("bpi", {}).items():
            prices.append(Price(d, Decimal(str(v))))

        return dataclasses.replace(series, prices=prices)

    def _data(self, series):
        url = "https://api.coindesk.com/v1/bpi/historical/close.json"
        params = {
            "currency": series.quote,
            "start": series.start,
            "end": series.end,
        }

        try:
            response = self.log_curl(requests.get(url, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        code = response.status_code
        text = response.text
        if code == 404 and "currency was not found" in text:
            raise exceptions.InvalidPair(series.base, series.quote, self)
        elif code == 404 and "only covers data from" in text:
            raise exceptions.BadResponse(text)
        elif code == 404 and "end date is before" in text and series.end < series.start:
            raise exceptions.BadResponse("End date is before start date.")
        elif code == 404 and "end date is before" in text:
            raise exceptions.BadResponse("The start date must be in the past.")
        elif code == 500 and "No results returned from database" in text:
            raise exceptions.BadResponse(
                "No results returned from database. This can happen when data "
                "for a valid quote currency (e.g. CUP) doesn't go all the way "
                "back to the start date, and potentially for other reasons."
            )
        else:
            try:
                response.raise_for_status()
            except Exception as e:
                raise exceptions.BadResponse(str(e)) from e

        try:
            result = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        return result



import re
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from typing import List, NamedTuple, Optional

from pricehist import exceptions
from pricehist.series import Series

SourcePrice = NamedTuple(
    "SourcePrice",
    [
        ("price", Decimal),
        ("time", Optional[datetime]),
        ("quote_currency", Optional[str]),
    ],
)


def source(pricehist_source):
    class Source:
        def get_latest_price(self, ticker: str) -> Optional[SourcePrice]:
            time_end = datetime.combine(date.today(), datetime.min.time())
            time_begin = time_end - timedelta(days=7)
            prices = self.get_prices_series(ticker, time_begin, time_end)
            if prices:
                return prices[-1]
            else:
                return None

        def get_historical_price(
            self, ticker: str, time: datetime
        ) -> Optional[SourcePrice]:
            prices = self.get_prices_series(ticker, time, time)
            if prices:
                return prices[-1]
            else:
                return None

        def get_prices_series(
            self,
            ticker: str,
            time_begin: datetime,
            time_end: datetime,
        ) -> Optional[List[SourcePrice]]:
            base, quote, type = self._decode(ticker)

            start = time_begin.date().isoformat()
            end = time_end.date().isoformat()

            local_tz = datetime.now(timezone.utc).astimezone().tzinfo
            user_tz = time_begin.tzinfo or local_tz

            try:
                series = pricehist_source.fetch(Series(base, quote, type, start, end))
            except exceptions.SourceError:
                return None

            return [
                SourcePrice(
                    price.amount,
                    datetime.fromisoformat(price.date).replace(tzinfo=user_tz),
                    series.quote,
                )
                for price in series.prices
            ]

        def _decode(self, ticker):
            # https://github.com/beancount/beanprice/blob/b05203/beanprice/price.py#L166
            parts = [
                re.sub(r"_[0-9a-fA-F]{2}", lambda m: chr(int(m.group(0)[1:], 16)), part)
                for part in ticker.split(":")
            ]
            base, quote, candidate_type = (parts + [""] * 3)[0:3]
            type = candidate_type or pricehist_source.types()[0]
            return (base, quote, type)

    return Source



from dataclasses import dataclass


@dataclass(frozen=True)
class Format:
    base: str = None
    quote: str = None
    time: str = "00:00:00"
    decimal: str = "."
    thousands: str = ""
    symbol: str = "rightspace"
    datesep: str = "-"
    csvdelim: str = ","
    jsonnums: bool = False

    @classmethod
    def fromargs(cls, args):
        def if_not_none(value, default):
            return default if value is None else value

        default = cls()
        return cls(
            base=if_not_none(args.formatbase, default.base),
            quote=if_not_none(args.formatquote, default.quote),
            time=if_not_none(args.formattime, default.time),
            decimal=if_not_none(args.formatdecimal, default.decimal),
            thousands=if_not_none(args.formatthousands, default.thousands),
            symbol=if_not_none(args.formatsymbol, default.symbol),
            datesep=if_not_none(args.formatdatesep, default.datesep),
            csvdelim=if_not_none(args.formatcsvdelim, default.csvdelim),
            jsonnums=if_not_none(args.formatjsonnums, default.jsonnums),
        )

    def format_date(self, date):
        return str(date).replace("-", self.datesep)

    def format_quote_amount(self, quote, amount):
        formatted_amount = self.format_num(amount)

        if self.symbol == "left":
            qa_parts = [quote, formatted_amount]
        elif self.symbol == "leftspace":
            qa_parts = [quote, " ", formatted_amount]
        elif self.symbol == "right":
            qa_parts = [formatted_amount, quote]
        else:
            qa_parts = [formatted_amount, " ", quote]

        quote_amount = "".join(qa_parts)

        return quote_amount

    def format_num(self, num):
        parts = f"{num:,}".split(".")
        parts[0] = parts[0].replace(",", self.thousands)
        result = self.decimal.join(parts)
        return result



"""
JSON output

Date, number and base/quote formatting options will be respected.

Classes:

    JSON

"""

import io
import json

from pricehist.format import Format

from .baseoutput import BaseOutput


class JSON(BaseOutput):
    def __init__(self, jsonl=False):
        self.jsonl = jsonl

    def format(self, series, source, fmt=Format()):
        data = []
        output = io.StringIO()

        base = fmt.base or series.base
        quote = fmt.quote or series.quote

        for price in series.prices:
            date = fmt.format_date(price.date)
            if fmt.jsonnums:
                amount = float(price.amount)
            else:
                amount = fmt.format_num(price.amount)

            data.append(
                {
                    "date": date,
                    "base": base,
                    "quote": quote,
                    "amount": amount,
                    "source": source.id(),
                    "type": series.type,
                }
            )

        if self.jsonl:
            for row in data:
                json.dump(row, output, ensure_ascii=False)
                output.write("\n")
        else:
            json.dump(data, output, ensure_ascii=False, indent=2)
            output.write("\n")

        return output.getvalue()



from dataclasses import dataclass, field, replace
from decimal import Decimal, getcontext
from typing import List

from pricehist.price import Price


@dataclass(frozen=True)
class Series:
    base: str
    quote: str
    type: str
    start: str
    end: str
    prices: List[Price] = field(default_factory=list)

    def invert(self):
        return replace(
            self,
            base=self.quote,
            quote=self.base,
            prices=[Price(date=p.date, amount=(1 / p.amount)) for p in self.prices],
        )

    def quantize(self, decimal_places):
        return replace(
            self,
            prices=[
                replace(p, amount=self._quantize(p.amount, decimal_places))
                for p in self.prices
            ],
        )

    def rename_base(self, new_base):
        return replace(self, base=new_base)

    def rename_quote(self, new_quote):
        return replace(self, quote=new_quote)

    def _quantize(self, amount, decimal_places):
        digits = len(amount.as_tuple().digits)
        exponent = amount.as_tuple().exponent

        fractional_digits = -exponent
        whole_digits = digits - fractional_digits
        max_decimal_places = getcontext().prec - whole_digits

        chosen_decimal_places = min(decimal_places, max_decimal_places)
        rounding = Decimal("0." + ("0" * chosen_decimal_places))

        return amount.quantize(rounding)



"""
GnuCash SQL output

Support for the `GnuCash <https://www.gnucash.org/>`_ accounting program is
achieved by generating SQL that can later be applied to a GnuCash database.

This allows pricehist to support GnuCash with simple text output rather than by
depending on GnuCash Python bindings or direct database interaction.

The generated SQL can be run in SQLite, MariaDB/MySQL or PostgreSQL.

Rows in GnuCash's prices table must include GUIDs for the related commodities.
The generated SQL selects the relevant GUIDs by mnemonic from the commodities
table and stores them in a temporary table. Another temprary table is populated
with new price data and the two are joined to produce the new rows that are
inserted into the prices table.

Users need to ensure that the base and quote of the new prices already have
commodities with matching mnemonics in the GnuCash database. If this condition
is not met, the SQL will fail without making changes. The names of the base and
quote can be adjusted with pricehist formatting options in case the source and
GnuCash names don't already match. Other formatting options can adjust date
formatting and the time of day used.

Each row in the prices table has a GUID of its own. These are generated in
pricehist by hashing the price data, so the same GUID will always be used for a
given date, base, quote, source, type & amount. Existing GUIDs are skipped
during the final insert into the prices table, so there's no problem with
running one SQL file multiple times or running multiple SQL files with
overlapping data.

Warnings are generated when string escaping or number limit issues are detected
and it should be easy for users to avoid those issues.

Classes:

    GnuCashSQL

"""

import hashlib
import logging
from datetime import datetime
from decimal import Decimal
from importlib.resources import read_text

from pricehist import __version__
from pricehist.format import Format

from .baseoutput import BaseOutput


class GnuCashSQL(BaseOutput):
    def format(self, series, source, fmt=Format()):
        base = fmt.base or series.base
        quote = fmt.quote or series.quote
        src = source.id()

        self._warn_about_backslashes(
            {
                "date": fmt.format_date("1970-01-01"),
                "time": fmt.time,
                "base": base,
                "quote": quote,
                "source": src,
                "price type": series.type,
            }
        )

        too_big = False
        values_parts = []
        for price in series.prices:
            date = f"{fmt.format_date(price.date)} {fmt.time}"
            m = hashlib.sha256()
            m.update(
                "".join(
                    [
                        date,
                        base,
                        quote,
                        src,
                        series.type,
                        str(price.amount),
                    ]
                ).encode("utf-8")
            )
            guid = m.hexdigest()[0:32]

            value_num, value_denom, fit = self._rational(price.amount)
            too_big |= not fit
            v = (
                "("
                + ", ".join(
                    [
                        self._sql_str(guid),
                        self._sql_str(date),
                        self._sql_str(base),
                        self._sql_str(quote),
                        self._sql_str(src),
                        self._sql_str(series.type),
                        str(value_num),
                        str(value_denom),
                    ]
                )
                + ")"
            )
            values_parts.append(v)
        values = ",\n".join(values_parts)
        values_comment = "" if values_parts else "-- "

        if too_big:
            # https://code.gnucash.org/docs/MAINT/group__Numeric.html
            # https://code.gnucash.org/docs/MAINT/structgnc__price__s.html
            logging.warning(
                "This SQL contains numbers outside of the int64 range required "
                "by GnuCash for the numerators and denominators of prices. "
                "Using the --quantize option to limit the number of decimal "
                "places will usually reduce the size of the rational form as "
                "well."
            )

        sql = read_text("pricehist.resources", "gnucash.sql").format(
            version=__version__,
            timestamp=datetime.utcnow().isoformat() + "Z",
            base=self._sql_str(base),
            quote=self._sql_str(quote),
            values_comment=values_comment,
            values=values,
        )

        return sql

    def _warn_about_backslashes(self, fields):
        hits = [name for name, value in fields.items() if "\\" in value]
        if hits:
            logging.warning(
                f"Before running this SQL, check the formatting of the "
                f"{self._english_join(hits)} strings. "
                f"SQLite treats backslashes in strings as plain characters, but "
                f"MariaDB/MySQL and PostgreSQL may interpret them as escape "
                f"codes."
            )

    def _english_join(self, strings):
        if len(strings) == 0:
            return ""
        elif len(strings) == 1:
            return str(strings[0])
        else:
            return f"{', '.join(strings[0:-1])} and {strings[-1]}"

    def _sql_str(self, s):
        # Documentation regarding SQL string literals:
        # * https://www.sqlite.org/lang_expr.html#literal_values_constants_
        # * https://mariadb.com/kb/en/string-literals/
        # * https://dev.mysql.com/doc/refman/8.0/en/string-literals.html
        # * https://www.postgresql.org/docs/devel/sql-syntax-lexical.html
        escaped = s.replace("'", "''")
        quoted = f"'{escaped}'"
        return quoted

    def _rational(self, number: Decimal) -> (str, str, bool):
        tup = number.as_tuple()
        sign = "-" if tup.sign == 1 else ""
        if tup.exponent > 0:
            numerator = (
                sign + "".join([str(d) for d in tup.digits]) + ("0" * tup.exponent)
            )
            denom = str(1)
        else:
            numerator = sign + "".join([str(d) for d in tup.digits])
            denom = str(10**-tup.exponent)
        fit = self._fit_in_int64(Decimal(numerator), Decimal(denom))
        return (numerator, denom, fit)

    def _fit_in_int64(self, *numbers):
        return all(n >= -(2**63) and n <= (2**63) - 1 for n in numbers)



"""
Beancount output

Supports the `Beancount <https://beancount.github.io/>`_ plain text accounting
format.

The default output should be valid for Beancount. Customizing it via formatting
options may generate invalid output, so users should keep the requirements of
the Beancount format in mind.

Relevant sections of the Beancount documentation:

* `Commodities / Currencies
   <https://beancount.github.io/docs/beancount_language_syntax.html#commodities-currencies>`_
* `Prices <https://beancount.github.io/docs/beancount_language_syntax.html#prices>`_
* `Fetching Prices in Beancount
   <https://beancount.github.io/docs/fetching_prices_in_beancount.html>`_

Classes:

    Beancount

"""

from pricehist.format import Format

from .baseoutput import BaseOutput


class Beancount(BaseOutput):
    def format(self, series, source=None, fmt=Format()):
        output = ""
        for price in series.prices:
            date = fmt.format_date(price.date)
            base = fmt.base or series.base
            quote = fmt.quote or series.quote
            quote_amount = fmt.format_quote_amount(quote, price.amount)
            output += f"{date} price {base} {quote_amount}\n"
        return output



import dataclasses
import json
from datetime import datetime, timedelta, timezone
from decimal import Decimal

import requests

from pricehist import exceptions
from pricehist.price import Price

from .basesource import BaseSource


class CoinbasePro(BaseSource):
    def id(self):
        return "coinbasepro"

    def name(self):
        return "Coinbase Pro"

    def description(self):
        return "The Coinbase Pro feed API provides market data to the public."

    def source_url(self):
        return "https://docs.pro.coinbase.com/"

    def start(self):
        return "2015-07-20"

    def types(self):
        return ["mid", "open", "high", "low", "close"]

    def notes(self):
        return (
            "This source uses Coinbase's Pro APIs, not the v2 API.\n"
            "No key or other authentication is requried because it only uses "
            "the feed APIs that provide market data and are public."
        )

    def symbols(self):
        products_url = "https://api.pro.coinbase.com/products"
        currencies_url = "https://api.pro.coinbase.com/currencies"

        try:
            products_response = self.log_curl(requests.get(products_url))
            currencies_response = self.log_curl(requests.get(currencies_url))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            products_response.raise_for_status()
            currencies_response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            products_data = json.loads(products_response.content)
            currencies_data = json.loads(currencies_response.content)
            currencies = {c["id"]: c for c in currencies_data}

            results = []
            for i in sorted(products_data, key=lambda i: i["id"]):
                base = i["base_currency"]
                quote = i["quote_currency"]
                base_name = currencies[base]["name"] if currencies[base] else base
                quote_name = currencies[quote]["name"] if currencies[quote] else quote
                results.append((f"{base}/{quote}", f"{base_name} against {quote_name}"))

        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if not results:
            raise exceptions.ResponseParsingError("Expected data not found")
        else:
            return results

    def fetch(self, series):
        data = []
        for seg_start, seg_end in self._segments(series.start, series.end):
            data.extend(self._data(series.base, series.quote, seg_start, seg_end))

        prices = []
        for item in data:
            prices.append(Price(item["date"], self._amount(item, series.type)))

        return dataclasses.replace(series, prices=prices)

    def _segments(self, start, end, length=290):
        start = datetime.fromisoformat(start).date()
        end = max(datetime.fromisoformat(end).date(), start)

        segments = []
        seg_start = start
        while seg_start <= end:
            seg_end = min(seg_start + timedelta(days=length - 1), end)
            segments.append((seg_start.isoformat(), seg_end.isoformat()))
            seg_start = seg_end + timedelta(days=1)

        return segments

    def _data(self, base, quote, start, end):
        product = f"{base}-{quote}"
        url = f"https://api.pro.coinbase.com/products/{product}/candles"
        params = {
            "start": start,
            "end": end,
            "granularity": "86400",
        }

        try:
            response = self.log_curl(requests.get(url, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        code = response.status_code
        text = response.text
        if code == 400 and "aggregations requested exceeds" in text:
            raise exceptions.BadResponse("Too many data points requested.")
        elif code == 400 and "start must be before end" in text:
            raise exceptions.BadResponse("The end can't preceed the start.")
        elif code == 400 and "is too old" in text:
            raise exceptions.BadResponse("The requested interval is too early.")
        elif code == 404 and "NotFound" in text:
            raise exceptions.InvalidPair(base, quote, self)
        elif code == 429:
            raise exceptions.RateLimit(
                "The rate limit has been exceeded. For more information see "
                "https://docs.pro.coinbase.com/#rate-limit."
            )
        else:
            try:
                response.raise_for_status()
            except Exception as e:
                raise exceptions.BadResponse(str(e)) from e

        try:
            result = reversed(
                [
                    {
                        "date": self._ts_to_date(candle[0]),
                        "low": candle[1],
                        "high": candle[2],
                        "open": candle[3],
                        "close": candle[4],
                    }
                    for candle in json.loads(response.content)
                    if start <= self._ts_to_date(candle[0]) <= end
                ]
            )
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        return result

    def _ts_to_date(self, ts):
        return datetime.fromtimestamp(ts, tz=timezone.utc).date().isoformat()

    def _amount(self, item, type):
        if type in ["mid"]:
            high = Decimal(str(item["high"]))
            low = Decimal(str(item["low"]))
            return sum([high, low]) / 2
        else:
            return Decimal(str(item[type]))



"""
CSV output

Comma Separated Values output is easily processed with other command-line tools
or imported into a spreadsheet or database.

Python's `csv <https://docs.python.org/3/library/csv.html>`_ module is used to
produce Excel-style CSV output, except with UNIX-style line endings. The field
delimiter can be set with a formatting option, and date, number and base/quote
formatting options will be respected.

Classes:

    CSV

"""

import csv
import io

from pricehist.format import Format

from .baseoutput import BaseOutput


class CSV(BaseOutput):
    def format(self, series, source, fmt=Format()):
        output = io.StringIO()
        writer = csv.writer(
            output,
            delimiter=fmt.csvdelim,
            lineterminator="\n",
            quotechar='"',
            doublequote=True,
            skipinitialspace=False,
            quoting=csv.QUOTE_MINIMAL,
        )

        header = ["date", "base", "quote", "amount", "source", "type"]
        writer.writerow(header)

        base = fmt.base or series.base
        quote = fmt.quote or series.quote

        for price in series.prices:
            date = fmt.format_date(price.date)
            amount = fmt.format_num(price.amount)
            writer.writerow([date, base, quote, amount, source.id(), series.type])

        return output.getvalue()



import logging
from datetime import date, datetime, timedelta

from pricehist import exceptions


def fetch(series, source, output, invert: bool, quantize: int, fmt) -> str:
    if series.start < source.start():
        logging.warning(
            f"The start date {series.start} preceeds the {source.name()} "
            f"source start date of {source.start()}."
        )

    with exceptions.handler():
        series = source.fetch(series)

    if len(series.prices) == 0:
        logging.warning(
            f"No data found for the interval [{series.start}--{series.end}]."
        )
    else:
        first = series.prices[0].date
        last = series.prices[-1].date
        message = (
            f"Available data covers the interval [{first}--{last}], "
            f"{_cov_description(series.start, series.end, first, last)}."
        )
        if first > series.start or last < series.end:
            expected_end = _yesterday() if series.end == _today() else series.end
            if first == series.start and last == expected_end:
                logging.debug(message)  # Missing today's price is expected
            else:
                logging.warning(message)
        else:
            logging.debug(message)

    if invert:
        series = series.invert()
    if quantize is not None:
        series = series.quantize(quantize)

    return output.format(series, source, fmt=fmt)


def _today():
    return date.today().isoformat()


def _yesterday():
    return (date.today() - timedelta(days=1)).isoformat()


def _cov_description(
    requested_start: str, requested_end: str, actual_start: str, actual_end: str
) -> str:
    date_format = "%Y-%m-%d"
    r1 = datetime.strptime(requested_start, date_format).date()
    r2 = datetime.strptime(requested_end, date_format).date()
    a1 = datetime.strptime(actual_start, date_format).date()
    a2 = datetime.strptime(actual_end, date_format).date()
    start_uncovered = (a1 - r1).days
    end_uncovered = (r2 - a2).days

    def s(n):
        return "" if n == 1 else "s"

    if start_uncovered == 0 and end_uncovered > 0:
        return (
            f"which ends {end_uncovered} day{s(end_uncovered)} earlier than "
            f"requested"
        )
    elif start_uncovered > 0 and end_uncovered == 0:
        return (
            f"which starts {start_uncovered} day{s(start_uncovered)} later "
            "than requested"
        )
    elif start_uncovered > 0 and end_uncovered > 0:
        return (
            f"which starts {start_uncovered} day{s(start_uncovered)} later "
            f"and ends {end_uncovered} day{s(end_uncovered)} earlier "
            f"than requested"
        )
    else:
        return "as requested"



import argparse
import logging
import shutil
import sys
from datetime import datetime, timedelta

from pricehist import __version__, logger, outputs, sources
from pricehist.fetch import fetch
from pricehist.format import Format
from pricehist.series import Series


def cli(argv=sys.argv):
    start_time = datetime.now()

    logger.init()

    parser = build_parser()
    args = parser.parse_args(argv[1:])

    if args.verbose:
        logger.show_debug()

    logging.debug(f"Began pricehist run at {start_time}.")

    try:
        if args.version:
            print(f"pricehist {__version__}")
        elif args.command == "sources":
            result = sources.formatted()
            print(result)
        elif args.command == "source" and args.symbols:
            result = sources.by_id[args.source].format_symbols()
            print(result, end="")
        elif args.command == "source" and args.search:
            result = sources.by_id[args.source].format_search(args.search)
            print(result, end="")
        elif args.command == "source":
            total_width = shutil.get_terminal_size().columns
            result = sources.by_id[args.source].format_info(total_width)
            print(result)
        elif args.command == "fetch":
            source = sources.by_id[args.source]
            output = outputs.by_type[args.output]
            if args.end < args.start:
                parser.error(
                    f"The end date '{args.end}' preceeds the start date '{args.start}'!"
                )
            if args.type not in source.types():
                parser.error(
                    f"The requested price type '{args.type}' is not "
                    f"recognized by the {source.id()} source!"
                )
            series = Series(
                base=source.normalizesymbol(args.pair[0]),
                quote=source.normalizesymbol(args.pair[1]),
                type=args.type,
                start=args.start,
                end=args.end,
            )
            fmt = Format.fromargs(args)
            result = fetch(series, source, output, args.invert, args.quantize, fmt)
            print(result, end="")
        else:
            parser.print_help()
    except BrokenPipeError:
        logging.debug("The output pipe was closed early.")
    finally:
        logging.debug(f"Ended pricehist run at {datetime.now()}.")


def valid_pair(s):
    base, quote = (s + "/").split("/")[0:2]
    if base == "":
        msg = f"No base found in the requested pair '{s}'."
        raise argparse.ArgumentTypeError(msg)
    return (base, quote)


def valid_date(s):
    if s == "today":
        return today()
    try:
        return datetime.strptime(s, "%Y-%m-%d").date().isoformat()
    except ValueError:
        msg = f"Not a valid YYYY-MM-DD date: '{s}'."
        raise argparse.ArgumentTypeError(msg)


def valid_date_before(s):
    return (
        datetime.strptime(valid_date(s), "%Y-%m-%d").date() - timedelta(days=1)
    ).isoformat()


def valid_date_after(s):
    return (
        datetime.strptime(valid_date(s), "%Y-%m-%d").date() + timedelta(days=1)
    ).isoformat()


def valid_char(s):
    if len(s) == 1:
        return s
    else:
        msg = f"Not a single character: '{s}'."
        raise argparse.ArgumentTypeError(msg)


def today():
    return datetime.now().date().isoformat()


def build_parser():
    def formatter(prog):
        return argparse.HelpFormatter(prog, max_help_position=50)

    class SetSourceDefaults(argparse.Action):
        def __call__(self, parser, namespace, value, option_string=None):
            source = sources.by_id[value]
            setattr(namespace, self.dest, value)
            if getattr(namespace, "type") is None:
                setattr(namespace, "type", source.types()[0])
            if getattr(namespace, "start") is None:
                setattr(namespace, "start", source.start())

    default_fmt = Format()
    parser = argparse.ArgumentParser(
        prog="pricehist",
        description="Fetch historical price data",
        formatter_class=formatter,
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="show version information",
    )

    parser.add_argument(
        "-vvv",
        "--verbose",
        action="store_true",
        help="show all log messages",
    )

    subparsers = parser.add_subparsers(title="commands", dest="command")

    sources_parser = subparsers.add_parser(
        "sources",
        help="list sources",
        formatter_class=formatter,
    )
    sources_parser.add_argument(
        "-vvv",
        "--verbose",
        action="store_true",
        help="show all log messages",
    )

    source_parser = subparsers.add_parser(
        "source",
        help="show source details",
        usage="pricehist source SOURCE [-h] [-s | --search QUERY]",
        formatter_class=formatter,
    )
    source_parser.add_argument(
        "source",
        metavar="SOURCE",
        type=str,
        choices=sources.by_id.keys(),
        help="the source identifier",
    )
    source_parser.add_argument(
        "-vvv",
        "--verbose",
        action="store_true",
        help="show all log messages",
    )

    source_list_or_search = source_parser.add_mutually_exclusive_group(required=False)
    source_list_or_search.add_argument(
        "-s",
        "--symbols",
        action="store_true",
        help="list available symbols",
    )
    source_list_or_search.add_argument(
        "--search",
        metavar="QUERY",
        type=str,
        help="search for symbols, if possible",
    )

    fetch_parser = subparsers.add_parser(
        "fetch",
        help="fetch prices",
        usage=(
            # Set usage manually to have positional arguments before options
            # and show allowed values where appropriate
            "pricehist fetch SOURCE PAIR [-h] [-vvv] "
            "[-t TYPE] [-s DATE | -sx DATE] [-e DATE | -ex DATE] "
            f"[-o {'|'.join(outputs.by_type.keys())}] "
            "[--invert] [--quantize INT] "
            "[--fmt-base SYM] [--fmt-quote SYM] [--fmt-time TIME] "
            "[--fmt-decimal CHAR] [--fmt-thousands CHAR] "
            "[--fmt-symbol rightspace|right|leftspace|left] [--fmt-datesep CHAR] "
            "[--fmt-csvdelim CHAR] [--fmt-jsonnums]"
        ),
        formatter_class=formatter,
    )
    fetch_parser.add_argument(
        "source",
        metavar="SOURCE",
        type=str,
        choices=sources.by_id.keys(),
        action=SetSourceDefaults,
        help="the source identifier",
    )
    fetch_parser.add_argument(
        "pair",
        metavar="PAIR",
        type=valid_pair,
        help="pair, usually BASE/QUOTE, e.g. BTC/USD",
    )
    fetch_parser.add_argument(
        "-vvv",
        "--verbose",
        action="store_true",
        help="show all log messages",
    )
    fetch_parser.add_argument(
        "-t",
        "--type",
        dest="type",
        metavar="TYPE",
        type=str,
        help="price type, e.g. close (default: first for source)",
    )
    fetch_start_group = fetch_parser.add_mutually_exclusive_group(required=False)
    fetch_start_group.add_argument(
        "-s",
        "--start",
        dest="start",
        metavar="DATE",
        type=valid_date,
        help="start date, inclusive (default: source start)",
    )
    fetch_start_group.add_argument(
        "-sx",
        "--startx",
        dest="start",
        metavar="DATE",
        type=valid_date_after,
        help="start date, exclusive",
    )

    fetch_end_group = fetch_parser.add_mutually_exclusive_group(required=False)
    fetch_end_group.add_argument(
        "-e",
        "--end",
        dest="end",
        metavar="DATE",
        type=valid_date,
        default=today(),
        help="end date, inclusive (default: today)",
    )
    fetch_end_group.add_argument(
        "-ex",
        "--endx",
        dest="end",
        metavar="DATE",
        type=valid_date_before,
        help="end date, exclusive",
    )

    fetch_parser.add_argument(
        "-o",
        "--output",
        dest="output",
        metavar="FMT",
        type=str,
        choices=outputs.by_type.keys(),
        default=outputs.default,
        help=f"output format (default: {outputs.default})",
    )
    fetch_parser.add_argument(
        "--invert",
        action="store_true",
        help="invert the price, swapping base and quote",
    )
    fetch_parser.add_argument(
        "--quantize",
        dest="quantize",
        metavar="INT",
        type=int,
        help="round to the given number of decimal places",
    )
    fetch_parser.add_argument(
        "--fmt-base",
        dest="formatbase",
        metavar="SYM",
        type=str,
        help="rename the base symbol in output",
    )
    fetch_parser.add_argument(
        "--fmt-quote",
        dest="formatquote",
        metavar="SYM",
        type=str,
        help="rename the quote symbol in output",
    )
    fetch_parser.add_argument(
        "--fmt-time",
        dest="formattime",
        metavar="TIME",
        type=str,
        help=f"set a particular time of day in output (default: {default_fmt.time})",
    )
    fetch_parser.add_argument(
        "--fmt-decimal",
        dest="formatdecimal",
        metavar="CHAR",
        type=str,
        help=f"decimal point in output (default: '{default_fmt.decimal}')",
    )
    fetch_parser.add_argument(
        "--fmt-thousands",
        dest="formatthousands",
        metavar="CHAR",
        type=str,
        help=f"thousands separator in output (default: '{default_fmt.thousands}')",
    )
    fetch_parser.add_argument(
        "--fmt-symbol",
        dest="formatsymbol",
        metavar="LOCATION",
        type=str,
        choices=["rightspace", "right", "leftspace", "left"],
        help=f"commodity symbol placement in output (default: {default_fmt.symbol})",
    )
    fetch_parser.add_argument(
        "--fmt-datesep",
        dest="formatdatesep",
        metavar="CHAR",
        type=str,
        help=f"date separator in output (default: '{default_fmt.datesep}')",
    )
    fetch_parser.add_argument(
        "--fmt-csvdelim",
        dest="formatcsvdelim",
        metavar="CHAR",
        type=valid_char,
        help=f"field delimiter for CSV output (default: '{default_fmt.csvdelim}')",
    )
    fetch_parser.add_argument(
        "--fmt-jsonnums",
        dest="formatjsonnums",
        action="store_true",
        help=f"numbers not strings for JSON output (default: {default_fmt.jsonnums})",
    )

    return parser



import csv
import dataclasses
import json
import logging
import os
from datetime import datetime, timedelta
from decimal import Decimal
from typing import List, Tuple

import requests

from pricehist import __version__, exceptions
from pricehist.price import Price

from .basesource import BaseSource


class AlphaVantage(BaseSource):
    QUERY_URL = "https://www.alphavantage.co/query"
    API_KEY_NAME = "ALPHAVANTAGE_API_KEY"

    def id(self):
        return "alphavantage"

    def name(self):
        return "Alpha Vantage"

    def description(self):
        return "Provider of market data for stocks, forex and cryptocurrencies"

    def source_url(self):
        return "https://www.alphavantage.co/"

    def start(self):
        return "1995-01-01"

    def types(self):
        return ["close", "open", "high", "low", "adjclose", "mid"]

    def notes(self):
        keystatus = "already set" if self._apikey(require=False) else "not yet set"
        return (
            "Alpha Vantage has data on digital (crypto) currencies, physical "
            "(fiat) currencies and stocks.\n"
            "You should obtain a free API key from "
            "https://www.alphavantage.co/support/#api-key and set it in "
            f"the {self.API_KEY_NAME} environment variable ({keystatus}), "
            "otherwise, pricehist will attempt to use a generic key.\n"
            "The PAIR for currencies should be in BASE/QUOTE form. The quote "
            "symbol must always be for a physical currency. The --symbols option "
            "will list all digital and physical currency symbols.\n"
            "The PAIR for stocks is the stock symbol only. The quote currency "
            f"will be determined automatically. {self._stock_symbols_message()}\n"
            "The price type 'adjclose' is only available for stocks.\n"
            "Beware that digital currencies quoted in non-USD currencies may "
            "be converted from USD data at one recent exchange rate rather "
            "than using historical rates.\n"
            "Alpha Vantage's standard API call frequency limits is 5 calls per "
            "minute and 500 per day, so you may need to pause between successive "
            "commands. Note that retrieving prices for one stock consumes two "
            "API calls."
        )

    def _stock_symbols_message(self):
        return "Stock symbols can be discovered using the --search option."

    def symbols(self):
        logging.info(self._stock_symbols_message())
        return self._digital_symbols() + self._physical_symbols()

    def search(self, query):
        data = self._search_data(query)
        results = [
            (
                m["1. symbol"],
                ", ".join(
                    [
                        m["2. name"],
                        m["3. type"],
                        m["4. region"],
                        m["8. currency"],
                    ]
                ),
            )
            for m in data["bestMatches"]
        ]
        return results

    def fetch(self, series):
        output_base = series.base.upper()
        output_quote = series.quote

        if series.quote == "":
            output_quote, data = self._stock_data(series)
        else:
            if series.type == "adjclose":
                raise exceptions.InvalidType(
                    series.type, series.base, series.quote, self
                )

            physical_symbols = [s for s, n in self._physical_symbols()]

            if series.quote not in physical_symbols:
                raise exceptions.InvalidPair(
                    series.base,
                    series.quote,
                    self,
                    "When given, the quote must be a physical currency.",
                )

            if series.base in physical_symbols:
                data = self._physical_data(series)

            elif series.base in [s for s, n in self._digital_symbols()]:
                data = self._digital_data(series)

            else:
                raise exceptions.InvalidPair(
                    series.base,
                    series.quote,
                    self,
                    "When a quote currency is given, the base must be a known "
                    "physical or digital currency.",
                )

        prices = [
            Price(day, amount)
            for day, entries in data.items()
            if (amount := self._amount(day, entries, series))
        ]

        return dataclasses.replace(
            series, base=output_base, quote=output_quote, prices=prices
        )

    def _amount(self, day, entries, series):
        if day < series.start or day > series.end:
            return None
        elif series.type == "mid":
            return sum([Decimal(entries["high"]), Decimal(entries["low"])]) / 2
        else:
            return Decimal(entries[series.type])

    def _stock_currency(self, symbol):
        data = self._search_data(symbol)
        for match in data["bestMatches"]:
            if match["1. symbol"] == symbol:
                return match["8. currency"]
        return None

    def _search_data(self, keywords: str):
        params = {
            "function": "SYMBOL_SEARCH",
            "keywords": keywords,
            "apikey": self._apikey(),
        }

        try:
            response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            data = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        self._raise_for_generic_errors(data)

        expected_keys = ["1. symbol", "2. name", "3. type", "4. region", "8. currency"]
        if (
            type(data) != dict
            or "bestMatches" not in data
            or type(data["bestMatches"]) != list
            or not all(k in m for k in expected_keys for m in data["bestMatches"])
        ):
            raise exceptions.ResponseParsingError("Unexpected content.")

        return data

    def _stock_data(self, series):
        output_quote = self._stock_currency(series.base) or "UNKNOWN"

        # As of 2022-11-24 TIME_SERIES_DAILY_ADJUSTED is no longer premium, but
        # now TIME_SERIES_DAILY is. So, always use TIME_SERIES_DAILY_ADJUSTED.
        function = "TIME_SERIES_DAILY_ADJUSTED"

        params = {
            "function": function,
            "symbol": series.base,
            "outputsize": self._outputsize(series.start),
            "apikey": self._apikey(),
        }

        try:
            response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            data = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        self._raise_for_generic_errors(data)

        if "Error Message" in data:
            if output_quote == "UNKNOWN":
                raise exceptions.InvalidPair(
                    series.base, series.quote, self, "Unknown stock symbol."
                )
            else:
                raise exceptions.BadResponse(data["Error Message"])

        try:
            normalized_data = {
                day: {
                    "open": entries["1. open"],
                    "high": entries["2. high"],
                    "low": entries["3. low"],
                    "close": entries["4. close"],
                    "adjclose": "5. adjusted close" in entries
                    and entries["5. adjusted close"],
                }
                for day, entries in reversed(data["Time Series (Daily)"].items())
            }
        except Exception as e:
            raise exceptions.ResponseParsingError("Unexpected content.") from e

        return output_quote, normalized_data

    def _physical_data(self, series):
        params = {
            "function": "FX_DAILY",
            "from_symbol": series.base,
            "to_symbol": series.quote,
            "outputsize": self._outputsize(series.start),
            "apikey": self._apikey(),
        }

        try:
            response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            data = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        self._raise_for_generic_errors(data)

        if type(data) != dict or "Time Series FX (Daily)" not in data:
            raise exceptions.ResponseParsingError("Unexpected content.")

        normalized_data = {
            day: {k[3:]: v for k, v in entries.items()}
            for day, entries in reversed(data["Time Series FX (Daily)"].items())
        }
        return normalized_data

    def _outputsize(self, start):
        almost_100_days_ago = (datetime.now().date() - timedelta(days=95)).isoformat()
        if start < almost_100_days_ago:
            return "full"
        else:
            return "compact"

    def _digital_data(self, series):
        params = {
            "function": "DIGITAL_CURRENCY_DAILY",
            "symbol": series.base,
            "market": series.quote,
            "apikey": self._apikey(),
        }

        try:
            response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            data = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        self._raise_for_generic_errors(data)

        if type(data) != dict or "Time Series (Digital Currency Daily)" not in data:
            raise exceptions.ResponseParsingError("Unexpected content.")

        normalized_data = {
            day: {
                "open": entries[f"1a. open ({series.quote})"],
                "high": entries[f"2a. high ({series.quote})"],
                "low": entries[f"3a. low ({series.quote})"],
                "close": entries[f"4a. close ({series.quote})"],
            }
            for day, entries in reversed(
                data["Time Series (Digital Currency Daily)"].items()
            )
        }
        return normalized_data

    def _apikey(self, require=True):
        key = os.getenv(self.API_KEY_NAME)
        if require and not key:
            generic_key = f"pricehist_{__version__}"
            logging.debug(
                f"{self.API_KEY_NAME} not set. "
                f"Defaulting to generic key '{generic_key}'."
            )
            return generic_key
        return key

    def _raise_for_generic_errors(self, data):
        if type(data) == dict:
            if "Note" in data and "call frequency" in data["Note"]:
                raise exceptions.RateLimit(data["Note"])
            if (
                "Information" in data
                and "unlock all premium endpoints" in data["Information"]
            ):
                msg = "You were denied access to a premium endpoint."
                raise exceptions.CredentialsError([self.API_KEY_NAME], self, msg)
            if "Error Message" in data and "apikey " in data["Error Message"]:
                raise exceptions.CredentialsError([self.API_KEY_NAME], self)

    def _physical_symbols(self) -> List[Tuple[str, str]]:
        url = "https://www.alphavantage.co/physical_currency_list/"
        return self._get_symbols(url, "Physical: ")

    def _digital_symbols(self) -> List[Tuple[str, str]]:
        url = "https://www.alphavantage.co/digital_currency_list/"
        return self._get_symbols(url, "Digital: ")

    def _get_symbols(self, url, prefix) -> List[Tuple[str, str]]:
        try:
            response = self.log_curl(requests.get(url))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            lines = response.content.decode("utf-8").splitlines()
            data = csv.reader(lines[1:], delimiter=",")
            results = [(s, f"{prefix}{n}") for s, n in data]
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if len(results) == 0:
            raise exceptions.ResponseParsingError("Symbols data missing.")

        return results



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pricehist',
 'pricehist.beanprice',
 'pricehist.outputs',
 'pricehist.resources',
 'pricehist.sources']

package_data = \
{'': ['*']}

install_requires = \
['cssselect>=1.1.0,<2.0.0',
 'curlify>=2.2.1,<3.0.0',
 'lxml>=4.6.2,<5.0.0',
 'requests>=2.25.1,<3.0.0']

entry_points = \
{'console_scripts': ['pricehist = pricehist.cli:cli']}

setup_kwargs = {
    'name': 'pricehist',
    'version': '1.4.3',
    'description': 'Fetch and format historical price data',
    'long_description': '# pricehist\n\nA command-line tool for fetching and formatting historical price data, with\nsupport for multiple data sources and output formats.\n\n[![Pipeline status](https://gitlab.com/chrisberkhout/pricehist/badges/master/pipeline.svg)](https://gitlab.com/chrisberkhout/pricehist/-/commits/master)\n[![Coverage report](https://gitlab.com/chrisberkhout/pricehist/badges/master/coverage.svg)](https://gitlab.com/chrisberkhout/pricehist/-/commits/master)\n[![PyPI version](https://badge.fury.io/py/pricehist.svg)](https://badge.fury.io/py/pricehist)\n[![Downloads](https://pepy.tech/badge/pricehist)](https://pepy.tech/project/pricehist)\n[![License](https://img.shields.io/pypi/l/pricehist)](https://gitlab.com/chrisberkhout/pricehist/-/blob/master/LICENSE)\n[![Code style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Hits](https://hits.seeyoufarm.com/api/count/incr/badge.svg?url=https%3A%2F%2Fgitlab.com%2Fchrisberkhout%2Fpricehist&count_bg=%2379C83D&title_bg=%23555555&icon=&icon_color=%23E7E7E7&title=hits&edge_flat=false)](https://hits.seeyoufarm.com)\n\n## Installation\n\nInstall via pip or\n[pipx](https://pypa.github.io/pipx/).\n\n```\npipx install pricehist\n```\n\n## Sources\n\n- **`alphavantage`**: [Alpha Vantage](https://www.alphavantage.co/)\n- **`bankofcanada`**: [Bank of Canada daily exchange rates](https://www.bankofcanada.ca/valet/docs)\n- **`coinbasepro`**: [Coinbase Pro](https://pro.coinbase.com/)\n- **`coindesk`**: [CoinDesk Bitcoin Price Index](https://www.coindesk.com/coindesk-api)\n- **`coinmarketcap`**: [CoinMarketCap](https://coinmarketcap.com/)\n- **`ecb`**: [European Central Bank Euro foreign exchange reference rates](https://www.ecb.europa.eu/stats/exchange/eurofxref/html/index.en.html)\n- **`yahoo`**: [Yahoo! Finance](https://finance.yahoo.com/)\n\n## Output formats\n\n- **`beancount`**: [Beancount](http://furius.ca/beancount/)\n- **`csv`**: [Comma-separated values](https://en.wikipedia.org/wiki/Comma-separated_values)\n- **`json`**: [JSON](https://en.wikipedia.org/wiki/JSON)\n- **`jsonl`**: [JSON lines](https://en.wikipedia.org/wiki/JSON_streaming)\n- **`gnucash-sql`**: [GnuCash](https://www.gnucash.org/) SQL\n- **`ledger`**: [Ledger](https://www.ledger-cli.org/) and [hledger](https://hledger.org/)\n\n## Reactions\n\n> This is my new favourite price fetcher, by far.  \n> -- _Simon Michael, creator of [hledger](https://hledger.org/) ([ref](https://groups.google.com/g/hledger/c/SCLbNiKl9D8/m/0ReYmDppAAAJ))_\n\n> This is great!  \n> -- _Martin Blais, creator of [Beancount](https://beancount.github.io/) ([ref](https://groups.google.com/g/beancount/c/cCJc9OhIlNg/m/QGRvNowcAwAJ))_\n\n## How to\n\n### Fetch prices\n\nFetch prices by choosing a source, a pair and, optionally, a time interval.\n\n```\npricehist fetch ecb EUR/AUD -s 2021-01-04 -e 2021-01-08\n```\n```\ndate,base,quote,amount,source,type\n2021-01-04,EUR,AUD,1.5928,ecb,reference\n2021-01-05,EUR,AUD,1.5927,ecb,reference\n2021-01-06,EUR,AUD,1.5824,ecb,reference\n2021-01-07,EUR,AUD,1.5836,ecb,reference\n2021-01-08,EUR,AUD,1.5758,ecb,reference\n```\n\nThe default output format is CSV, which is suitable for use in spreadsheets and\nwith other tools. For example, you can generate a price chart from the command\nline as follows (or using [an alias](https://gitlab.com/-/snippets/2163031)).\n\n```\npricehist fetch coindesk BTC/USD -s 2021-01-01 | \\\n  sed 1d | \\\n  cut -d, -f1,4 | \\\n  gnuplot -p -e \'\n    set datafile separator ",";\n    set xdata time;\n    set timefmt "%Y-%m-%d";\n    set format x "%b\\n%Y";\n    plot "/dev/stdin" using 1:2 with lines title "BTC/USD"\n  \'\n```\n\n![BTC/USD prices](https://gitlab.com/chrisberkhout/pricehist/-/raw/master/example-gnuplot.png)\n\n### Show usage information\n\nAdd `-h` to any command to see usage information.\n\n```\npricehist fetch -h\n```\n```\nusage: pricehist fetch SOURCE PAIR [-h] [-vvv] [-t TYPE] [-s DATE | -sx DATE] [-e DATE | -ex DATE]\n[-o beancount|csv|json|jsonl|gnucash-sql|ledger] [--invert] [--quantize INT]\n[--fmt-base SYM] [--fmt-quote SYM] [--fmt-time TIME] [--fmt-decimal CHAR] [--fmt-thousands CHAR]\n[--fmt-symbol rightspace|right|leftspace|left] [--fmt-datesep CHAR]\n[--fmt-csvdelim CHAR] [--fmt-jsonnums]\n\npositional arguments:\n  SOURCE                   the source identifier\n  PAIR                     pair, usually BASE/QUOTE, e.g. BTC/USD\n\noptional arguments:\n  -h, --help               show this help message and exit\n  -vvv, --verbose          show all log messages\n  -t TYPE, --type TYPE     price type, e.g. close\n  -s DATE, --start DATE    start date, inclusive (default: source start)\n  -sx DATE, --startx DATE  start date, exclusive\n  -e DATE, --end DATE      end date, inclusive (default: today)\n  -ex DATE, --endx DATE    end date, exclusive\n  -o FMT, --output FMT     output format (default: csv)\n  --invert                 invert the price, swapping base and quote\n  --quantize INT           round to the given number of decimal places\n  --fmt-base SYM           rename the base symbol in output\n  --fmt-quote SYM          rename the quote symbol in output\n  --fmt-time TIME          set a particular time of day in output (default: 00:00:00)\n  --fmt-decimal CHAR       decimal point in output (default: \'.\')\n  --fmt-thousands CHAR     thousands separator in output (default: \'\')\n  --fmt-symbol LOCATION    commodity symbol placement in output (default: rightspace)\n  --fmt-datesep CHAR       date separator in output (default: \'-\')\n  --fmt-csvdelim CHAR      field delimiter for CSV output (default: \',\')\n  --fmt-jsonnums           numbers not strings for JSON output (default: False)\n```\n\n### Choose and customize the output format\n\nAs the output format you can choose one of `beancount`, `csv`, `json`, `jsonl`,\n`ledger` or `gnucash-sql`.\n\n```\npricehist fetch ecb EUR/AUD -s 2021-01-04 -e 2021-01-08 -o ledger\n```\n```\nP 2021-01-04 00:00:00 EUR 1.5928 AUD\nP 2021-01-05 00:00:00 EUR 1.5927 AUD\nP 2021-01-06 00:00:00 EUR 1.5824 AUD\nP 2021-01-07 00:00:00 EUR 1.5836 AUD\nP 2021-01-08 00:00:00 EUR 1.5758 AUD\n```\n\nFormatting options let you control certain details of the output.\n\n```\npricehist fetch ecb EUR/AUD -s 2021-01-04 -e 2021-01-08 -o ledger \\\n  --fmt-time \'\' --fmt-datesep / --fmt-base € --fmt-quote $ --fmt-symbol left\n```\n```\nP 2021/01/04 € $1.5928\nP 2021/01/05 € $1.5927\nP 2021/01/06 € $1.5824\nP 2021/01/07 € $1.5836\nP 2021/01/08 € $1.5758\n```\n\n### Fetch new prices only\n\nYou can update an existing file without refetching the prices you already have.\nFirst find the date of the last price, then fetch from there, drop the header\nline if present and append the rest to the existing file.\n\n```\nlast=$(tail -1 prices-eur-usd.csv | cut -d, -f1)\npricehist fetch ecb EUR/USD -sx $last -o csv | sed 1d >> prices-eur-usd.csv\n```\n\n### Load prices into GnuCash\n\nYou can generate SQL for a GnuCash database and apply it immediately with one\nof the following commands.\n\n```\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o gnucash-sql | sqlite3 Accounts.gnucash\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o gnucash-sql | mysql -u username -p -D databasename\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o gnucash-sql | psql -U username -d databasename -v ON_ERROR_STOP=1\n```\n\nBeware that the GnuCash project itself does not support integration at the\ndatabase level, so there is a risk that the SQL generated by `pricehist` will\nbe ineffective or even damaging for some version of GnuCash. In practice, this\nstrategy has been used successfully by other projects. Reading the SQL and\nkeeping regular database backups is recommended.\n\nThe GnuCash database must already contain commodities with mnemonics matching\nthe base and quote of new prices, otherwise the SQL will fail without making\nchanges.\n\nEach price entry is given a GUID based on its content (date, base, quote,\nsource, type and amount) and existing GUIDs are skipped in the final insert, so\nyou can apply identical or overlapping SQL files multiple times without\ncreating duplicate entries in the database.\n\n### Show source information\n\nThe `source` command shows information about a source.\n\n```\npricehist source alphavantage\n```\n```\nID          : alphavantage\nName        : Alpha Vantage\nDescription : Provider of market data for stocks, forex and cryptocurrencies\nURL         : https://www.alphavantage.co/\nStart       : 1995-01-01\nTypes       : close, open, high, low, adjclose, mid\nNotes       : Alpha Vantage has data on...\n```\n\nAvailable symbols can be listed for most sources, either as full pairs or as\nseparate base and quote symbols that will work in certain combinations.\n\n```\npricehist source ecb --symbols\n```\n```\nEUR/AUD    Euro against Australian Dollar\nEUR/BGN    Euro against Bulgarian Lev\nEUR/BRL    Euro against Brazilian Real\nEUR/CAD    Euro against Canadian Dollar\nEUR/CHF    Euro against Swiss Franc\n...\n```\n\nIt may also be possible to search for symbols.\n\n```\npricehist source alphavantage --search Tesla\n```\n```\nTL0.DEX       Tesla, Equity, XETRA, EUR\nTL0.FRK       Tesla, Equity, Frankfurt, EUR\nTSLA34.SAO    Tesla, Equity, Brazil/Sao Paolo, BRL\nTSLA          Tesla Inc, Equity, United States, USD\nTXLZF         Tesla Exploration Ltd, Equity, United States, USD\n```\n\n### Inspect source interactions\n\nYou can see extra information by adding the verbose option (`--verbose` or\n`-vvv`), including `curl` commands that reproduce each request to a source.\n\n```\npricehist fetch coindesk BTC/USD -s 2021-01-01 -e 2021-01-05 -vvv\n```\n```\nDEBUG Began pricehist run at 2021-08-12 14:38:26.630357.\nDEBUG Starting new HTTPS connection (1): api.coindesk.com:443\nDEBUG https://api.coindesk.com:443 "GET /v1/bpi/historical/close.json?currency=USD&start=2021-01-01&end=2021-01-05 HTTP/1.1" 200 319\nDEBUG curl -X GET -H \'Accept: */*\' -H \'Accept-Encoding: gzip, deflate\' -H \'Connection: keep-alive\' -H \'User-Agent: python-requests/2.25.1\' --compressed \'https://api.coindesk.com/v1/bpi/historical/close.json?currency=USD&start=2021-01-01&end=2021-01-05\'\nDEBUG Available data covers the interval [2021-01-01--2021-01-05], as requested.\ndate,base,quote,amount,source,type\n2021-01-01,BTC,USD,29391.775,coindesk,close\n2021-01-02,BTC,USD,32198.48,coindesk,close\n2021-01-03,BTC,USD,33033.62,coindesk,close\n2021-01-04,BTC,USD,32017.565,coindesk,close\n2021-01-05,BTC,USD,34035.0067,coindesk,close\nDEBUG Ended pricehist run at 2021-08-12 14:38:26.709428.\n```\n\nRunning a logged `curl` command shows exactly what data is returned by the\nsource.\n\n```\npricehist fetch coindesk BTC/USD -s 2021-01-01 -e 2021-01-05 -vvv 2>&1 \\\n  | grep \'^DEBUG curl\' | sed \'s/^DEBUG //\' | bash | jq .\n```\n```json\n{\n  "bpi": {\n    "2021-01-01": 29391.775,\n    "2021-01-02": 32198.48,\n    "2021-01-03": 33033.62,\n    "2021-01-04": 32017.565,\n    "2021-01-05": 34035.0067\n  },\n  "disclaimer": "This data was produced from the CoinDesk Bitcoin Price Index. BPI value data returned as USD.",\n  "time": {\n    "updated": "Jan 6, 2021 00:03:00 UTC",\n    "updatedISO": "2021-01-06T00:03:00+00:00"\n  }\n}\n```\n\n### Use via `bean-price`\n\nBeancount users may wish to use `pricehist` sources via `bean-price`. To do so,\nensure the `pricehist` package is installed in an accessible location.\n\nYou can fetch the latest price directly from the command line.\n\n```\nbean-price -e "USD:pricehist.beanprice.coindesk/BTC:USD"\n```\n```\n2021-08-18 price BTC:USD                          44725.12 USD\n```\n\nYou can fetch a series of prices by providing a Beancount file as input.\n\n```\n; input.beancount\n2021-08-14 commodity BTC\n  price: "USD:pricehist.beanprice.coindesk/BTC:USD:close"\n```\n\n```\nbean-price input.beancount --update --update-rate daily --inactive --clear-cache\n```\n```\n2021-08-14 price BTC                            47098.2633 USD\n2021-08-15 price BTC                            47018.9017 USD\n2021-08-16 price BTC                             45927.405 USD\n2021-08-17 price BTC                            44686.3333 USD\n2021-08-18 price BTC                              44725.12 USD\n```\n\nAdding `-v` will print progress information, `-vv` will print debug information,\nincluding that from `pricehist`.\n\nA source map specification for `bean-price` has the form\n`<currency>:<module>/[^]<ticker>`. Additional `<module>/[^]<ticker>` parts can\nbe appended, separated by commas.\n\nThe module name will be of the form `pricehist.beanprice.<source_id>`.\n\nThe ticker symbol will be of the form `BASE:QUOTE:TYPE`.\n\nAny non-alphanumeric characters except the equals sign (`=`), hyphen (`-`),\nperiod (`.`), or parentheses (`(` or `)`) are special characters that need to\nbe encoded as their a two-digit hexadecimal code prefixed with an underscore,\nbecause `bean-price` ticker symbols don\'t allow all the characters used by\n`pricehist` pairs.\n[This page](https://replit.com/@chrisberkhout/bpticker) will do it for you.\n\nFor example, the Yahoo! Finance symbol for the Dow Jones Industrial Average is\n`^DJI`, and would have the source map specification\n`USD:pricehist.beanprice.yahoo/_5eDJI`, or for the daily high price\n`USD:pricehist.beanprice.yahoo/_5eDJI::high`.\n\n### Use as a library\n\nYou may find `pricehist`\'s source classes useful in your own scripts.\n\n```\n$\xa0python\nPython 3.9.6 (default, Jun 30 2021, 10:22:16)\n[GCC 11.1.0] on linux\nType "help", "copyright", "credits" or "license" for more information.\n>>> from pricehist.series import Series\n>>> from pricehist.sources.ecb import ECB\n>>> series = ECB().fetch(Series("EUR", "AUD", "reference", "2021-01-04", "2021-01-08"))\n>>> series.prices\n[Price(date=\'2021-01-04\', amount=Decimal(\'1.5928\')), Price(date=\'2021-01-05\', amount=Decimal(\'1.5927\')), Price(date=\'2021-01-06\', amount=Decimal(\'1.5824\')), Price(date=\'2021-01-07\', amount=Decimal(\'1.5836\')), Price(date=\'2021-01-08\', amount=Decimal(\'1.5758\'))]\n```\n\nA subclass of `pricehist.exceptions.SourceError` will be raised for any error.\n\n### Contribute\n\nContributions are welcome! If you discover a bug or want to work on a\nnon-trivial change, please open a\n[GitLab issue](https://gitlab.com/chrisberkhout/pricehist/-/issues)\nto discuss it.\n\nRun `make install-pre-commit-hook` set up local pre-commit checks.\nSet up your editor to run\n[isort](https://pycqa.github.io/isort/),\n[Black](https://black.readthedocs.io/en/stable/) and\n[Flake8](https://flake8.pycqa.org/en/latest/),\nor run them manually via `make format lint`.\n\n## Terminology\n\nA **source** is an upstream service that can provide a series of prices.\n\nEach **series** of prices is for one pair and price type.\n\nThe [**pair**](https://en.wikipedia.org/wiki/Currency_pair) is made up of a\nbase and a quote, each given as a symbol. Sometimes you will give the base\nonly, and the quote will be determined with information from the source. The\navailable pairs, the symbols used in them and the available price types all\ndepend on the particular source used.\n\nThe **base** is the currency or commodity being valued. Each price expresses\nthe value of one unit of the base.\n\nThe **quote** is the unit used to express the value of the base.\n\nA **symbol** is a code or abbreviation for a currency or commodity.\n\nThe **prices** in a series each have a date and an amount.\n\nThe **amount** is the number of units of the quote that are equal to one unit\nof the base.\n\nConsider the following command.\n\n```\npricehist fetch coindesk BTC/USD --type close\n```\n\n- **`coindesk`** is the ID of the CoinDesk Bitcoin Price Index source.\n- **`BTC`** is the symbol for Bitcoin, used here as the base.\n- **`USD`** is the symbol for the United States Dollar, used here as the quote.\n- **`BTC/USD`** is the pair Bitcoin against United States Dollar.\n- **`close`** is the price type for the last price of each day.\n\nA BTC/USD price of the amount 29,391.775 can be written as\n"BTC/USD = 29391.775" or "BTC 29391.775 USD", and means that one Bitcoin is\nworth 29,391.775 United States Dollars.\n\n## Initial design choices\n\nTo keep things simple, `pricehist` provides only univariate time series of\ndaily historical prices. It doesn\'t provide other types of market, financial or\neconomic data, real-time prices, or other temporal resolutions. Multiple or\nmultivariate series require multiple invocations.\n\n## Potential features\n\nIn the future, `pricehist` may be extended to cover some of the following\nfeatures:\n\n- **Time of day**: Sources sometimes provide specific times for each day\'s\n  high/low prices and these could be preserved for output. This would require\n  changes to how dates are handled internally, clarification of time zone\n  handling and extension of the time formatting option.\n- **Alternate resolutions**: Some sources can provide higher or lower\n  resolution data, such as hourly or weekly. These could be supported where\n  available. For other cases an option could be provided for downsampling data\n  before output.\n- **Real-time prices**: These generally come from different source endpoints\n  than the historical data. Real-time prices will usually have a different\n  price type, such as `last`, `bid` or `ask`. Support for real-time prices\n  would allow adding sources that don\'t provide historical data. Start and end\n  times are irrelevant when requesting real-time prices. A "follow" option\n  could continuously poll for new prices.\n- **Related non-price data**: Trading volume, spreads, split and dividend\n  events and other related data could be supported. The base/quote/type model\n  used for prices would work for some of this. Other things may require\n  extending the model.\n- **Multivariate series**: Would allow, for example, fetching\n  high/low/open/close prices in a single invocation.\n- **`format` command**: A command for rewriting existing CSV data into one of\n  the other output formats.\n\n## Alternatives\n\nBeancount\'s [`bean-price`](https://github.com/beancount/beanprice) tool fetches\nprices and addresses other workflow concerns in a Beancount-specific manner,\ngenerally requiring a Beancount file as input.\n\nThe [Piecash](https://piecash.readthedocs.io/) library is a pythonic interface\nto GnuCash files stored in SQL which has a\n[`Commodity.update_prices`](https://piecash.readthedocs.io/en/master/api/piecash.core.commodity.html?highlight=update_prices#piecash.core.commodity.Commodity.update_prices)\nmethod for fetching historical prices.\nThe GnuCash wiki documents [wrapper scripts](https://wiki.gnucash.org/wiki/Stocks/get_prices)\nfor the [Finance::QuoteHist](https://metacpan.org/pod/Finance::QuoteHist) Perl\nmodule.\n',
    'author': 'Chris Berkhout',
    'author_email': 'chris@chrisberkhout.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/chrisberkhout/pricehist',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import dataclasses
from datetime import datetime, timedelta
from decimal import Decimal

import requests
from lxml import etree

from pricehist import exceptions, isocurrencies
from pricehist.price import Price

from .basesource import BaseSource


class ECB(BaseSource):
    def id(self):
        return "ecb"

    def name(self):
        return "European Central Bank"

    def description(self):
        return "European Central Bank Euro foreign exchange reference rates"

    def source_url(self):
        return "https://www.ecb.europa.eu/stats/exchange/eurofxref/html/index.en.html"

    def start(self):
        return "1999-01-04"

    def types(self):
        return ["reference"]

    def notes(self):
        return ""

    def symbols(self):
        quotes = self._quotes()
        iso = isocurrencies.by_code()
        return [
            (f"EUR/{c}", f"Euro against {iso[c].name if c in iso else c}")
            for c in quotes
        ]

    def fetch(self, series):
        if series.base != "EUR" or not series.quote:  # EUR is the only valid base.
            raise exceptions.InvalidPair(series.base, series.quote, self)

        almost_90_days_ago = (datetime.now().date() - timedelta(days=85)).isoformat()
        root = self._data(series.start < almost_90_days_ago)

        all_rows = []
        for day in root.cssselect("[time]"):
            date = day.attrib["time"]
            for row in day.cssselect(f"[currency='{series.quote}']"):
                rate = Decimal(row.attrib["rate"])
                all_rows.insert(0, (date, rate))

        if not all_rows and series.quote not in self._quotes():
            raise exceptions.InvalidPair(series.base, series.quote, self)

        selected = [
            Price(d, r) for d, r in all_rows if d >= series.start and d <= series.end
        ]

        return dataclasses.replace(series, prices=selected)

    def _quotes(self):
        root = self._data(more_than_90_days=True)
        nodes = root.cssselect("[currency]")
        quotes = sorted(set([n.attrib["currency"] for n in nodes]))
        if not quotes:
            raise exceptions.ResponseParsingError("Expected data not found")
        return quotes

    def _data(self, more_than_90_days=False):
        url_base = "https://www.ecb.europa.eu/stats/eurofxref"
        if more_than_90_days:
            source_url = f"{url_base}/eurofxref-hist.xml"  # since 1999
        else:
            source_url = f"{url_base}/eurofxref-hist-90d.xml"  # last 90 days

        try:
            response = self.log_curl(requests.get(source_url))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            root = etree.fromstring(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        return root



"""
Ledger output

Supports both `Ledger <https://www.ledger-cli.org/>`_ and
`hledger <https://hledger.org/>`_ plain text accounting formats.

By default the output should be valid for Ledger, but can be customized for
hledger or other variants via formatting options. Invalid variants are
possible, so the user should be familiar with the requirements of the target
format.

Relevant sections of the Ledger manual:

* `Commodities and Currencies
   <https://www.ledger-cli.org/3.0/doc/ledger3.html#Commodities-and-Currencies>`_
* `Commoditized Amounts
   <https://www.ledger-cli.org/3.0/doc/ledger3.html#Commoditized-Amounts>`_

Relevant sections of the hledger manual:

* `Declaring market prices <https://hledger.org/hledger.html#declaring-market-prices>`_:
* `Declaring commodities <https://hledger.org/hledger.html#declaring-commodities`_:

Classes:

    Ledger

"""

from pricehist.format import Format

from .baseoutput import BaseOutput


class Ledger(BaseOutput):
    def format(self, series, source=None, fmt=Format()):
        output = ""
        for price in series.prices:
            date = fmt.format_date(price.date)
            base = fmt.base or series.base
            quote = fmt.quote or series.quote
            quote_amount = fmt.format_quote_amount(quote, price.amount)
            timesep = " " if fmt.time else ""
            output += f"P {date}{timesep}{fmt.time} {base} {quote_amount}\n"
        return output



import dataclasses
import json
from datetime import datetime, timezone
from decimal import Decimal

import requests

from pricehist import exceptions
from pricehist.price import Price

from .basesource import BaseSource


class CoinMarketCap(BaseSource):
    def id(self):
        return "coinmarketcap"

    def name(self):
        return "CoinMarketCap"

    def description(self):
        return "The world's most-referenced price-tracking website for cryptoassets"

    def source_url(self):
        return "https://coinmarketcap.com/"

    def start(self):
        return "2013-04-28"

    def types(self):
        return ["mid", "open", "high", "low", "close"]

    def notes(self):
        return (
            "This source makes unoffical use of endpoints that power CoinMarketCap's "
            "public web interface. The price data comes from a public equivalent of "
            "the OHLCV Historical endpoint found in CoinMarketCap's official API.\n"
            "CoinMarketCap currency symbols are not necessarily unique, so it "
            "is recommended that you use IDs, which can be listed via the "
            "--symbols option. For example, 'ETH/BTC' is 'id=1027/id=1'. The "
            "corresponding symbols will be used in output."
        )

    def symbols(self):
        data = self._symbol_data()
        ids = [f"id={i['id']}" for i in data]
        descriptions = [f"{i['symbol'] or i['code']} {i['name']}".strip() for i in data]
        return list(zip(ids, descriptions))

    def fetch(self, series):
        if series.base == "ID=" or not series.quote or series.quote == "ID=":
            raise exceptions.InvalidPair(series.base, series.quote, self)

        data = self._data(series)

        prices = []
        for item in data.get("quotes", []):
            d = item["time_open"][0:10]
            amount = self._amount(next(iter(item["quote"].values())), series.type)
            prices.append(Price(d, amount))

        output_base, output_quote = self._output_pair(series.base, series.quote, data)

        return dataclasses.replace(
            series, base=output_base, quote=output_quote, prices=prices
        )

    def _data(self, series):
        url = "https://web-api.coinmarketcap.com/v1/cryptocurrency/ohlcv/historical"

        params = {}

        if series.base.startswith("ID="):
            params["id"] = series.base[3:]
        else:
            params["symbol"] = series.base

        if series.quote.startswith("ID="):
            params["convert_id"] = series.quote[3:]
        else:
            params["convert"] = series.quote

        params["time_start"] = int(
            int(
                datetime.strptime(series.start, "%Y-%m-%d")
                .replace(tzinfo=timezone.utc)
                .timestamp()
            )
            - 24 * 60 * 60
            # Start one period earlier since the start is exclusive.
        )
        params["time_end"] = int(
            datetime.strptime(series.end, "%Y-%m-%d")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        )  # Don't round up since it's inclusive of the period covering the end time.

        try:
            response = self.log_curl(requests.get(url, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        code = response.status_code
        text = response.text

        if code == 400 and "No items found." in text:
            raise exceptions.InvalidPair(
                series.base, series.quote, self, "Bad base ID."
            )

        elif code == 400 and 'Invalid value for \\"convert_id\\"' in text:
            raise exceptions.InvalidPair(
                series.base, series.quote, self, "Bad quote ID."
            )

        elif code == 400 and 'Invalid value for \\"convert\\"' in text:
            raise exceptions.InvalidPair(
                series.base, series.quote, self, "Bad quote symbol."
            )

        elif code == 400 and "must be older than" in text:
            if series.start <= series.end:
                raise exceptions.BadResponse("The start date must be in the past.")
            else:
                raise exceptions.BadResponse(
                    "The start date must preceed or match the end date."
                )

        elif (
            code == 400
            and "must be a valid ISO 8601 timestamp or unix time" in text
            and series.start < "2001-09-11"
        ):
            raise exceptions.BadResponse("The start date can't preceed 2001-09-11.")

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            parsed = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if type(parsed) != dict or "data" not in parsed:
            raise exceptions.ResponseParsingError("Unexpected content.")

        elif len(parsed["data"]) == 0:
            raise exceptions.ResponseParsingError(
                "The data section was empty. This can happen when the quote "
                "currency symbol can't be found, and potentially for other reasons."
            )

        return parsed["data"]

    def _amount(self, data, type):
        if type in ["mid"]:
            high = Decimal(str(data["high"]))
            low = Decimal(str(data["low"]))
            return sum([high, low]) / 2
        else:
            return Decimal(str(data[type]))

    def _output_pair(self, base, quote, data):
        data_base = data["symbol"]

        data_quote = None
        if len(data["quotes"]) > 0:
            data_quote = next(iter(data["quotes"][0]["quote"].keys()))

        lookup_quote = None
        if quote.startswith("ID="):
            symbols = {i["id"]: (i["symbol"] or i["code"]) for i in self._symbol_data()}
            lookup_quote = symbols[int(quote[3:])]

        output_base = data_base
        output_quote = lookup_quote or data_quote or quote

        return (output_base, output_quote)

    def _symbol_data(self):
        base_url = "https://web-api.coinmarketcap.com/v1/"
        fiat_url = f"{base_url}fiat/map?include_metals=true"
        crypto_url = f"{base_url}cryptocurrency/map?sort=cmc_rank"

        fiat = self._get_json_data(fiat_url)
        crypto = self._get_json_data(crypto_url)

        return crypto + fiat

    def _get_json_data(self, url, params={}):
        try:
            response = self.log_curl(requests.get(url, params=params))
        except Exception as e:
            raise exceptions.RequestError(str(e)) from e

        try:
            response.raise_for_status()
        except Exception as e:
            raise exceptions.BadResponse(str(e)) from e

        try:
            parsed = json.loads(response.content)
        except Exception as e:
            raise exceptions.ResponseParsingError(str(e)) from e

        if type(parsed) != dict or "data" not in parsed:
            raise exceptions.ResponseParsingError("Unexpected content.")

        elif len(parsed["data"]) == 0:
            raise exceptions.ResponseParsingError("Empty data section.")

        return parsed["data"]
